import base64
import configparser
import logging
import os
from typing import Optional, Callable
from Crypto.Cipher import AES
from hashlib import md5

CONFIG_FOLDER = os.getenv('OS_CONFIG_FOLDER', '/etc')
CONFIG_FN = os.getenv('OS_CONFIG_FILENAME')

DEFAULT_CONFIG_FN = 'defaultAssetConfig.ini'
DEFAULT_CONFIG_KY = None
ASSET_CONFIGS = {}
RELOAD_FN = os.path.join(CONFIG_FOLDER, 'reload')


def get_key(agentcore_uuid):
    flipped = 'tt{0}'.format(agentcore_uuid[::-1]).encode('utf-8')
    return md5(flipped).hexdigest().encode('utf-8')


def unpad(s):
    return s[0:-bytearray(s)[-1]]


def decrypt(key, data):
    enc = base64.b64decode(data)
    iv = enc[:AES.block_size]
    cipher = AES.new(key, AES.MODE_CBC, iv)
    dec = cipher.decrypt(enc[AES.block_size:])
    return unpad(dec).decode('utf-8')


def get_asset_config(
        asset_id: str,
        ip4: Optional[str],
        agentcore_uuid: str,
        func: Callable) -> Optional[dict]:
    if os.path.exists(RELOAD_FN):
        ASSET_CONFIGS.clear()
        os.unlink(RELOAD_FN)

    cred = ASSET_CONFIGS.get(asset_id)
    if cred:
        return cred

    fn = os.path.join(CONFIG_FOLDER, f'{asset_id}.ini')
    if os.path.exists(fn):
        key = get_key(agentcore_uuid)
        config = configparser.ConfigParser()
        config.read(fn)
        try:
            ASSET_CONFIGS[asset_id] = cred = func(config, key, decrypt)
        except Exception as e:
            logging.error(f'Config `{fn}` error: {e}')
        return cred  # cred can be None in case of an error

    if ip4:
        cred = ASSET_CONFIGS.get(ip4)
        if cred:
            # make sure next time this will be found for asset_id
            ASSET_CONFIGS[asset_id] = cred
            return cred

        fn = os.path.join(CONFIG_FOLDER, f'{ip4}.ini')
        if os.path.exists(fn):
            key = get_key(agentcore_uuid)
            config = configparser.ConfigParser()
            config.read(fn)
            try:
                ASSET_CONFIGS[asset_id] = cred = func(config, key, decrypt)
            except Exception as e:
                logging.error(f'Config `{fn}` error: {e}')
            return cred  # cred can be None in case of an error

    cred = ASSET_CONFIGS.get(DEFAULT_CONFIG_KY)
    if cred:
        # make sure next time this will be found for asset_id
        ASSET_CONFIGS[asset_id] = cred
        return cred

    fn = os.path.join(CONFIG_FOLDER, DEFAULT_CONFIG_FN)
    if os.path.exists(fn):
        key = get_key(agentcore_uuid)
        config = configparser.ConfigParser()
        config.read(fn)
        try:
            ASSET_CONFIGS[DEFAULT_CONFIG_KY] = cred = \
                func(config, key, decrypt)
        except Exception as e:
            logging.error(f'Config `{fn}` error: {e}')
        else:
            return cred

    # Log with debug level because some probes just do not require secrets
    logging.debug(f'No default config file found ({fn})')
    ASSET_CONFIGS[DEFAULT_CONFIG_KY] = None



class IgnoreResultException(Exception):
    """IgnoreResultException should be raised by a check if the result needs
    to be ignored.
    Nothing for this check will be returned to the AgentCore.
    """
    pass



import asyncio
import json
import logging
import os
import platform
import socket
import sys
import time
import uuid
from typing import Optional, Callable

from .config import CONFIG_FN
from .config import get_asset_config
from .logger import setup_logger
from .protocol import Protocol
from .exceptions import IgnoreResultException


PROC_START_TS = int(time.time())
SYSTEM_ID = str(uuid.uuid1()).split('-')[-1]


class AgentCoreClient:

    def __init__(
        self,
        probe_name: str,
        version: str,
        checks: dict,
        read_asset_config: Optional[Callable] = None,
        config_fn: Optional[str] = None
    ):
        logging.info(f"starting {probe_name} v{version}")
        self._loop = asyncio.get_event_loop()
        self.connecting = False
        self.connected = False
        self._protocol = None
        self._keepalive = None
        self._on_asset_config = read_asset_config
        self._probe_name = probe_name
        self._probe_version = version
        self._checks = checks
        self._required_services = [
            k
            for k, c in checks.items()
            if getattr(c, 'required', False)
        ]

        config_fn = CONFIG_FN or config_fn

        config = json.load(open(config_fn)) if config_fn and \
            os.path.exists(config_fn) else {}

        logging.debug(f"config (source: {config_fn}): {config}")

        self.host = os.getenv(
            'OS_AGENTCORE_IP',
            config.get('agentCoreIp', 'localhost'))
        self.port = int(os.getenv(
            'OS_AGENTCORE_PORT',
            config.get('agentCorePort', 7211)))

        logging.debug(f"using agentcore: {self.host}:{self.port}")

    @staticmethod
    def setup_logger(log_level: str = 'warning', log_colorized: bool = False):
        setup_logger(log_level, log_colorized)

    async def _connect(self):
        conn = self._loop.create_connection(
            lambda: Protocol(
                self.on_connection_made,
                self.on_connection_lost,
                self.on_run_check,
            ),
            self.host,
            self.port
        )

        self.connecting = True
        try:
            _, self._protocol = await asyncio.wait_for(conn, timeout=10)
        except Exception as e:
            logging.error(f'connecting to agentcore failed: {e}')
        else:
            if self._keepalive is None or self._keepalive.done():
                self._keepalive = asyncio.ensure_future(self._keepalive_loop())
            self._announce()

        self.connecting = False

    async def _keepalive_loop(self):
        step = 30
        await asyncio.sleep(step)  # start with a delay

        while self.connected:
            try:
                self._protocol.send({'type': 'echoRequest'})
            except asyncio.CancelledError:
                break
            except Exception as e:
                logging.error(e)
                self.close()
                break
            await asyncio.sleep(step)

    async def connect_loop(self):
        initial_step = 2
        step = 2
        max_step = 2 ** 7

        while 1:
            if not self.connected and not self.connecting:
                await self._connect()
                step = min(step * 2, max_step)
            else:
                step = initial_step
            await asyncio.sleep(step)

    def close(self):
        if self._keepalive is not None:
            self._keepalive.cancel()
            self._keepalive = None
        if self._protocol is not None:
            self._protocol.transport.close()
            self._protocol = None

    def on_connection_made(self):
        logging.warning('connected to agentcore')
        self.connected = True

    def on_connection_lost(self):
        logging.error('connection to agentcore lost')
        self.connected = False

    def send(self, msg):
        if self._protocol and self._protocol.transport:
            self._protocol.send(msg)

    def _announce(self):
        self._protocol.send({
            'type': 'probeAnnouncement',
            'hostInfo': self._get_hostinfo(),
            'platform': self._get_platform_str(),
            'versionNr': self._probe_version,
            'probeName': self._probe_name,
            'probeProperties': ['remoteProbe'],
            'availableChecks': {
                k: {
                    'defaultCheckInterval': v.interval,
                    'requiredServices': self._required_services,
                } for k, v in self._checks.items()
            },
        })

    @staticmethod
    def _get_hostinfo():
        return {
            'timestamp': int(time.time()),
            'hostName': socket.getfqdn(),
            'osFamily': os.name,
            'platform': platform.system(),
            'ip4': socket.gethostbyname(socket.gethostname()),
            'release': platform.release(),
            'systemId': SYSTEM_ID,
            'processStartTs': PROC_START_TS
        }

    @staticmethod
    def _get_platform_str():
        platform_bits = 'x64' if sys.maxsize > 2 ** 32 else 'x32'
        return f'{platform.system()}_{platform_bits}_{platform.release()}'

    @staticmethod
    def _get_framework(
            check,
            check_name: str,
            start_time: float,
            is_available: bool):
        framework = {
            'timestamp': start_time,
            'runtime': time.time() - start_time
        }
        if getattr(check, 'required', False):
            msg = 'OK' if is_available else 'see check error(s) for details'
            framework['serviceInfo'] = {'services': {check_name: {
                'available': is_available,
                'reportedBy': check_name,
                'serviceInfoMsg': msg,
                'hasDependingChecks': True
            }}}
        return framework

    async def on_run_check(self, data):
        try:
            asset_id = data['hostUuid']
            asset_config = data['hostConfig']
            check_name = data['checkName']
            agentcore_uuid = asset_config['parentCore']
            config = asset_config['probeConfig'][self._probe_name]
            ip4 = config.get('ip4')
            check = self._checks[check_name]
        except Exception:
            logging.error('invalid check configuration')
            return

        cred: Optional[dict] = self._on_asset_config and get_asset_config(
            asset_id, ip4, agentcore_uuid, self._on_asset_config)

        t0 = time.time()
        try:
            state_data = await check.run(data, cred)
        except IgnoreResultException:
            logging.info(f'on_run_check {asset_id} {check_name} (ignored)')
        except Exception as e:
            logging.warning(f'on_run_check {asset_id} {check_name} {e}')
            message = str(e)
            framework = self._get_framework(
                check, check_name, t0, is_available=False)
            self.send({
                'type': 'checkError',
                'hostUuid': asset_id,
                'checkName': check_name,
                'message': message,
                'framework': framework,
            })
        else:
            if state_data:
                logging.debug(f'on_run_check {asset_id} {check_name} ok!')
                framework = self._get_framework(
                    check, check_name, t0, is_available=True)
                self.send({
                    'type': 'stateData',
                    'hostUuid': asset_id,
                    'framework': framework,
                    'checkName': check_name,
                    'stateData': state_data
                })



import colorlog
import logging.handlers
import os

_LOG_LEVEL = os.getenv('OS_LOG_LEVEL')

_LOG_DATE_FMT = '%y%m%d %H:%M:%S'

_MAP_LOG_LEVELS = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL
}


def setup_logger(log_level: str = 'warning', log_colorized: bool = False):
    """Setup logger.

    Keyword arguments:
        log_level: Log level to use. May be overwritten by the environment
                   variable `OS_LOG_LEVEL`. Default: `warning`.
        log_colors: Use colors for logging.
    """

    if log_colorized:
        # setup colorized formatter
        formatter = colorlog.ColoredFormatter(
            fmt=(
                '%(log_color)s[%(levelname)1.1s %(asctime)s %(module)s'
                ':%(lineno)d]%(reset)s %(message)s'),
            datefmt=_LOG_DATE_FMT,
            reset=True,
            log_colors={
                'DEBUG': 'cyan',
                'INFO': 'green',
                'WARNING': 'yellow',
                'ERROR': 'red',
                'CRITICAL': 'red,bg_white'},
            secondary_log_colors={},
            style='%')
    else:
        # setup formatter without using colors
        formatter = logging.Formatter(
            fmt=(
                '[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d] '
                '%(message)s'),
            datefmt=_LOG_DATE_FMT,
            style='%')

    logger = logging.getLogger()

    log_level = _LOG_LEVEL or log_level
    logger.setLevel(_MAP_LOG_LEVELS[log_level.upper()])

    ch = logging.StreamHandler()

    # we can set the handler level to DEBUG since we control the root level
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(formatter)
    logger.addHandler(ch)



try:
    from .client import AgentCoreClient
except ImportError:
    pass  # importing msgpack might fail when importing from setup.py

from .version import __version__
from .exceptions import IgnoreResultException
